/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.data;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import net.yale.cs.p4p.api.APIBase.HTTPErrorException;
import net.yale.cs.p4p.api.data.InetService;
import net.yale.cs.p4p.api.data.PID;
import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.portal.api.QueryREST;

import org.gudy.azureus2.plugins.peers.PeerDescriptor;

import se.kth.nada.router.RoutingTable;

public final class PeerCache extends QueryCache
{
	private static final Collection<InetAddress> EMPTY_ADDRS = new ArrayList<InetAddress>();

	public static final PID UNKNOWN_PID = new PID("none", 0, true);

	/* Background task used to resolve PIDs */
	private final class PIDQueryTask extends QueryTask {

		public PIDQueryTask(int interval) {
			super(interval);
		}

		private InetService discoverPortalServer() {
			try {
				String portalServerHost = getPlugin().getConfig().getPortalHost();
				if (portalServerHost != null && portalServerHost.length() > 0)
					return new InetService(InetAddress.getByName(portalServerHost), getPlugin().getConfig().getPortalPort());
			} catch (UnknownHostException e) {
				getPlugin().getLogger().log("Invalid user-specified Portal Server: " + e);
				return null;
			}

			try {
				return getPlugin().getPortalDiscoveryInterface().discover();
			} catch (HTTPErrorException e) {
				getPlugin().getLogger().log("Portal discovery failed: HTTP error " + e.getCode() + " " + e.getReason());
				return null;
			} catch (IOException e) {
				getPlugin().getLogger().log("Portal discovery failed: " + e);
				return null;
			}
		}

		private Map<InetAddress, PID> queryPIDs(InetService portalServer, Collection<InetAddress> ips) {
			try {
				return new QueryREST(portalServer).getPIDs(ips);
			} catch (IOException e) {
				getPlugin().getLogger().log("Failed to query PIDs: " + e);
				return null;
			}
		}

		private void doPortalDiscovery() {
			portalServer_ = discoverPortalServer();
			getPlugin().getLogger().log("Located Portal server " + portalServer_);
		}

		private void doSelfQuery() {
			if (portalServer_ == null)
				return;
			
			Map<InetAddress, PID> pids = queryPIDs(portalServer_, EMPTY_ADDRS);
			if (pids == null)
				return;

			Map.Entry<InetAddress, PID> e = pids.entrySet().iterator().next();
			if (getPlugin().getLogger().isEnabled())
				getPlugin().getLogger().log("External IP address is " + e.getKey() + " in PID " + e.getValue());

			synchronized (PeerCache.this) {
				selfPid_ = e.getValue();
				selfIp_ = e.getKey();
			}
		}

		private void doPeersQuery() {
			int maxQueryBatch = getPlugin().getConfig().getPIDQueryBatch();
			Collection<PeerDescriptor> peers = new ArrayList<PeerDescriptor>(maxQueryBatch);

			/* Get the next batch of peers to query */
			synchronized (PeerCache.this) {
				for (PeerDescriptor peer : pendingPeers_) {
					if (peers.size() >= maxQueryBatch)
						break;
					peers.add(peer);
				}
				pendingPeers_.removeAll(peers);
			}

			/* Quit early if nothing to query or if no portal server
			 * NOTE: This means that peers we just dequeued will be
			 *       entirely forgotten. */
			if (portalServer_ == null || peers.isEmpty()) {
				synchronized (PeerCache.this) {
					removedPeers_.clear();
				}
				return;
			}

			/* Extract IP addresses */
			Collection<InetAddress> addrs = new ArrayList<InetAddress>(peers.size());
			for (PeerDescriptor peer : peers) {
				try {
					addrs.add(InetAddress.getByName(peer.getIP()));
				} catch (UnknownHostException e) {
					getPlugin().getLogger().log("failed to convert " + peer.getIP() + " to IP address", e);
				}
			}

			/* Do the query */
			Map<InetAddress, PID> pids = queryPIDs(portalServer_, addrs);
			if (pids == null) {
				synchronized (PeerCache.this) {
					removedPeers_.clear();
				}
				return;
			}

			/* Convert resulting IP addresses into strings
			 * so we can use them for lookups */
			Map<String, PID> pidStrings = new HashMap<String, PID>(pids.size());
			for (Map.Entry<InetAddress, PID> e : pids.entrySet())
				pidStrings.put(e.getKey().getHostAddress(), e.getValue() != null ? e.getValue() : UNKNOWN_PID);
			
			synchronized (PeerCache.this) {
				for (PeerDescriptor peer : peers) {
					
					/* If this peer was removed, ignore it */
					if (removedPeers_.contains(peer))
						continue;

					addResolvedPeer(peer, pidStrings.get(peer.getIP()));
				}

				/* We've verified results against removed peers,
				 * so we no longer need them. */
				removedPeers_.clear();
			}
			
		}

		protected void query() throws InterruptedException {

			if (needPortalQuery())
				doPortalDiscovery();

			if (needSelfQuery())
				doSelfQuery();

			doPeersQuery();
			
			setInterval(getPlugin().getConfig().getPIDQueryInterval() * 1000);
		}
	}

	/* Peers not yet assigned to PIDs */
	private LinkedHashSet<PeerDescriptor> pendingPeers_ = new LinkedHashSet<PeerDescriptor>();
	
	/* All peers currently assigned PIDs */
	private Map<PeerDescriptor, PID> peersByPeer_ = new HashMap<PeerDescriptor, PID>();
	private Map<PID, Set<PeerDescriptor>> peersByPid_ = new HashMap<PID, Set<PeerDescriptor>>();

	/* Peers that have been recently removed. This list is
	 * checked and cleared in the background query task. */
	private Set<PeerDescriptor> removedPeers_ = new HashSet<PeerDescriptor>();

	/* Time of last portal discovery */
	private long lastPortalQuery_ = 0;
	private synchronized boolean needPortalQuery() {
		boolean result = lastPortalQuery_ + getPlugin().getConfig().getPortalDiscoverInterval() * 1000 < System.currentTimeMillis();
		lastPortalQuery_ = System.currentTimeMillis();
		return result;
	}

	/* Portal server */
	private InetService portalServer_;
	public InetService getPortalServer() { return portalServer_; }

	/* Time of last query for our own PID */
	private long lastSelfQuery_ = 0;
	private synchronized boolean needSelfQuery() {
		boolean result = lastSelfQuery_ + getPlugin().getConfig().getPIDQuerySelfInterval() * 1000 < System.currentTimeMillis();
		lastSelfQuery_ = System.currentTimeMillis();
		return result;
	}

	/* Our own PID; null indicates not yet queried */
	private PID selfPid_ = null;
	public PID getSelfPID() { return selfPid_; }

	/* Our own IP address; null indicates not yet queried */
	private InetAddress selfIp_ = null;
	public InetAddress getSelfIP() { return selfIp_; }

	public PeerCache(P4PPlugin plugin) {
		super(plugin, 1);
	}

	protected QueryTask makeQueryTask() {
		return new PIDQueryTask(getPlugin().getConfig().getPIDQueryInterval() * 1000);
	}

	private static PID lookupPID(RoutingTable pidMap, String ipAddr) {
		PID pid = pidMap.find(ipAddr);
		return (pid != null) ? pid : UNKNOWN_PID;
	}

	public synchronized PID lookupPID(String ipAddr) {
		RoutingTable pidMap = getPlugin().getPIDMapCache().getPIDMap();
		if (pidMap == null)
			return UNKNOWN_PID;
		
		return lookupPID(pidMap, ipAddr);
	}

	private void addResolvedPeer(PeerDescriptor peer, PID pid) {		
		if (getPlugin().getLogger().isEnabled())
			getPlugin().getLogger().log("Added peer " + peer.getIP() + " in PID " + pid);
		
		peersByPeer_.put(peer, pid);
		Set<PeerDescriptor> pidPeers = peersByPid_.get(pid);
		if (pidPeers == null)
			peersByPid_.put(pid, pidPeers = new HashSet<PeerDescriptor>());
		pidPeers.add(peer);
	}

	/* Returns true if resolved, false if they will be resolved in the background */
	public synchronized boolean addPeers(Collection<PeerDescriptor> peers) {

		RoutingTable pidMap = getPlugin().getPIDMapCache().getPIDMap();
		try {
			for (PeerDescriptor peer : peers) {
				/* Ignore if we already know about it */
				if (peersByPeer_.containsKey(peer))
					continue;
				
				if (pidMap != null)
					addResolvedPeer(peer, lookupPID(pidMap, peer.getIP()));
				else
					pendingPeers_.add(peer);
			}

			return pidMap != null;
		} finally {
			removedPeers_.removeAll(peers);
		}
	}

	public synchronized void removePeer(PeerDescriptor peer) {
		/* Remove current records of the peer */
		pendingPeers_.remove(peer);
		
		PID peerPid = peersByPeer_.remove(peer);
		Set<PeerDescriptor> pidPeers = peersByPid_.get(peerPid);
		if (pidPeers != null) {
			/* Remove record of the peer from the PID bucket */
			pidPeers.remove(peer);
			
			/* Also remove record of the PID if it is now empty */
			if (pidPeers.isEmpty())
				peersByPid_.remove(peerPid);
		}
		
		/* Allow removal of peers that could possibly
		 * be getting resolved in the background right now.
		 */
		removedPeers_.add(peer);
	}

	public synchronized PID getPID(PeerDescriptor peer) {
		return peersByPeer_.get(peer);
	}

	public synchronized Set<PID> getPIDs() {
		return new HashSet<PID>(peersByPid_.keySet());
	}

	public synchronized Set<PeerDescriptor> getPendingPeers() {
		return new HashSet<PeerDescriptor>(pendingPeers_);
	}

	public synchronized Set<PeerDescriptor> getPeers(PID pid) {
		Set<PeerDescriptor> peers = peersByPid_.get(pid);
		if (peers == null)
			return Collections.emptySet();
		return new HashSet<PeerDescriptor>(peers);
	}

}
