package net.yale.cs.p4p.p2p.azureus.client.data.selection;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import net.yale.cs.p4p.api.data.PID;
import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.PeerCache;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState.PeerSnapshot;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class PeerSelectionP4PSimple extends PeerSelectionBase {

	private static final ScheduledPeerInfo FALLBACK_BASE_SCHED_INFO = new ScheduledPeerInfo(Collections.singleton("alg=RANDOM"));
	private static final ScheduledPeerInfo P4P_BASE_SCHED_INFO = new ScheduledPeerInfo(Collections.singleton("alg=P4PSIMPLE"));

	public PeerSelectionP4PSimple(P4PPlugin plugin) {
		super(plugin);
	}

	protected ScheduleResult selectPeers(Download download, Torrent torrent, PeerManager peerManager, PeerSnapshot schedPeers, int count) {

		/* Ensure we know what our own PID is.  If we don't, revert to random selection */
		PID selfPid = getPeerCache().getSelfPID();
		if (selfPid == null) {
			getPlugin().getLogger().log(torrent, "No Self PID; reverting to random selection");
			return selectPeersRandom(schedPeers, count, FALLBACK_BASE_SCHED_INFO);
		}

		double intrapidPct = getPlugin().getConfig().getSimpleAlgIntrapidPct();
		double intradomainPct = getPlugin().getConfig().getSimpleAlgIntradomainPct();
		double interdomainPct = getPlugin().getConfig().getSimpleAlgInterdomainPct();

		/* Construct list of result peers */
		List<ScheduledPeer> selectedPeers = new ArrayList<ScheduledPeer>(count);

		/* Select from available peers */
		fillScheduledPeers(
			selectPeers(torrent, schedPeers.getAvailPeers(), selfPid, intrapidPct, intradomainPct, interdomainPct, count),
			selectedPeers,
			new ScheduledPeerInfo(P4P_BASE_SCHED_INFO, Collections.singleton("available")));
		
		return new ScheduleResult(selectedPeers);
	}

	protected Peer selectOptimisticUnchoke(Download download, Torrent torrent, Collection<Peer> peers) {

		/* If we don't have a PID, we don't select anyone */
		PID selfPid = getPeerCache().getSelfPID();
		if (selfPid == null)
			return null;

		/* Resolve peers into their corresponding PIDs.
		 * FIXME: We currently require PID Map to be available */
		ArrayList<Peer> intrapidPeers = new ArrayList<Peer>();
		ArrayList<Peer> intradomainPeers = new ArrayList<Peer>();
		ArrayList<Peer> interdomainPeers = new ArrayList<Peer>();

		/* Separate all peers into the three categories:
		 * intrapid: Peers within the same PID
		 * intradomain: Peers with PIDs marked as internal
		 * interdomain: Peers with PIDs marked as external, or UNKNOWN */
		for (Peer peer : peers) {
			PID pid = getPeerCache().lookupPID(peer.getIp());

			if (pid == null || pid.getExternal() || PeerCache.UNKNOWN_PID.equals(pid))
				interdomainPeers.add(peer);
			else if (selfPid.equals(pid))
				intrapidPeers.add(peer);
			else
				intradomainPeers.add(peer);
		}
		
		double intrapidPct = getPlugin().getConfig().getSimpleAlgIntrapidUnchokePct();
		double intradomainPct = getPlugin().getConfig().getSimpleAlgIntradomainUnchokePct();
		double interdomainPct = getPlugin().getConfig().getSimpleAlgInterdomainUnchokePct();

		double r = getRNG().nextDouble();
		if (r <= intrapidPct && !intrapidPeers.isEmpty())
			return getRandomElement(intrapidPeers);

		if (r <= intradomainPct && !intradomainPeers.isEmpty())
			return getRandomElement(intradomainPeers);

		if (r <= interdomainPct && !interdomainPeers.isEmpty())
			return getRandomElement(interdomainPeers);

		return null;
	}

	protected List<PeerDescriptor> selectPeers(Torrent torrent, Collection<PeerDescriptor> availPeers, PID selfPid, double intrapidPct, double intradomainPct, double interdomainPct, int count) {
		
		if (count <= 0)
			return Collections.emptyList();
		
		ArrayList<PeerDescriptor> intrapidPeers = new ArrayList<PeerDescriptor>();
		ArrayList<PeerDescriptor> intradomainPeers = new ArrayList<PeerDescriptor>();
		ArrayList<PeerDescriptor> interdomainPeers = new ArrayList<PeerDescriptor>();

		/* Separate all peers into the three categories:
		 * intrapid: Peers within the same PID
		 * intradomain: Peers with PIDs marked as internal
		 * interdomain: Peers with PIDs marked as external, or UNKNOWN */
		for (PeerDescriptor peer : availPeers) {
			PID pid = getPeerCache().getPID(peer);

			if (pid == null || pid.getExternal() || PeerCache.UNKNOWN_PID.equals(pid))
				interdomainPeers.add(peer);
			else if (selfPid.equals(pid))
				intrapidPeers.add(peer);
			else
				intradomainPeers.add(peer);
		}

		/* Construct list of result peers */
		List<PeerDescriptor> selectedPeers = new ArrayList<PeerDescriptor>(count);

		/* Add intrapid */
		selectedPeers.addAll(getRandomElements(intrapidPeers, Math.max(0, (int)(intrapidPct * count) - selectedPeers.size())));
		getPlugin().getLogger().log(torrent, "Selected total of " + selectedPeers.size() + " peers after intrapid stage");
		
		selectedPeers.addAll(getRandomElements(intradomainPeers, Math.max(0, (int)(intradomainPct * count) - selectedPeers.size())));
		getPlugin().getLogger().log(torrent, "Selected total of " + selectedPeers.size() + " peers after intradomain stage");

		selectedPeers.addAll(getRandomElements(interdomainPeers, Math.max(0, (int)(interdomainPct * count) - selectedPeers.size())));
		getPlugin().getLogger().log(torrent, "Selected total of " + selectedPeers.size() + " peers after interdomain stage");

		return selectedPeers;
	}

}
