package net.yale.cs.p4p.p2p.azureus.client.data.weights.opt;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;



public class OptProblemBase
{
	
	public static final class Variable implements Comparable<Variable> {
		
		private VariableSet parent_;
		public VariableSet getParent() { return parent_; }

		private int varSetIndex_;
		public int getVarSetIndex() { return varSetIndex_; }

		private Variable(VariableSet parent, int varSetIndex) {
			parent_ = parent;
			varSetIndex_ = varSetIndex;
		}
		
		public int getVarIdx() {
			return parent_.getBaseIndex() + getVarSetIndex();
		}

		public int hashCode() {
			return varSetIndex_;
		}
		
		public boolean equals(Object o) {
			if (o == this)
				return true;
			
			if (o == null || !(o instanceof Variable))
				return false;
			
			Variable oVar = (Variable)o;
			return varSetIndex_ == oVar.varSetIndex_ && parent_.equals(oVar.parent_);
		}

		public int compareTo(Variable o) {
			int rc = Integer.valueOf(varSetIndex_).compareTo(o.varSetIndex_);
			if (rc != 0)
				return rc;
			
			return parent_.compareTo(o.parent_);
		}
	}

	public static final class VariableSet implements Comparable<VariableSet> {

		private OptProblemBase parent_;
		public OptProblemBase getParent() { return parent_; }

		private String name_;
		public String getName() { return name_; }

		private int baseIdx_;
		public int getBaseIndex() { return baseIdx_; }
		
		private int n1_;
		private int n2_;

		private VariableSet(String name, int baseIdx, int n1, int n2) {
			baseIdx_ = baseIdx;
			n1_ = n1;
			n2_ = n2;
		}

		public Variable getVar(int i) {
			return getVar(i, 0);
		}

		public Variable getVar(int i, int j) {
			if (i >= n1_ || j >= n2_)
				throw new IllegalArgumentException("Variable index out of bounds");
			
			return new Variable(this, i * n2_ + j);
		}
		
		public int getVarCount() {
			return n1_ * n2_;
		}

		public int hashCode() {
			return name_.hashCode();
		}
		
		public boolean equals(Object o) {
			if (this == o)
				return true;
			
			if (o == null || !(o instanceof VariableSet))
				return false;

			VariableSet oSet = (VariableSet)o;
			
			return name_.equals(oSet.name_)
				&& baseIdx_ == oSet.baseIdx_
				&& n1_ == oSet.n1_
				&& n2_ == oSet.n2_;
		}

		public int compareTo(VariableSet o) {
			int rc = name_.compareTo(o.name_);
			if (rc != 0)
				return rc;
			
			rc = Integer.valueOf(baseIdx_).compareTo(o.baseIdx_);
			if (rc != 0)
				return rc;
			
			rc = Integer.valueOf(n1_).compareTo(o.n1_);
			if (rc != 0)
				return rc;
			
			return Integer.valueOf(n2_).compareTo(o.n2_);
		}
	}

	public static class Expression {
		
		private Map<Variable, Double> expr_ = new HashMap<Variable, Double>();
		public Map<Variable, Double> getExpr() { return Collections.unmodifiableMap(expr_); }
		public void addVar(Variable var) {
			addVar(var, 1);
		}
		public void addVar(Variable var, double value) {
			expr_.put(var, value);
		}
		
	}

	public static final class Constraint {
		
		public static enum Type {
			LESS,
			GREATER,
		}

		private Expression expr_ = new Expression();
		public Expression getExpr() { return expr_; }

		private Type type_ = Type.LESS;
		public Type getType() { return type_; }
		
		private double rhs_ = 0;
		public double getRHS() { return rhs_; }
		
		public Constraint(Type type, double rhs) {
			type_ = type;
			rhs_ = rhs;
		}
	}
	
	public static class ObjectiveBase {

		private double coeffs[];
		
		public ObjectiveBase(int varCount, Expression obj) {
			coeffs = new double[varCount];

			/* All coefficients initially 0 */
			for (int i = 0; i < coeffs.length; ++i)
				coeffs[i] = 0;
			
			for (Map.Entry<Variable, Double> e : obj.getExpr().entrySet())
				coeffs[e.getKey().getVarIdx()] = e.getValue();
		}

		public double function(double[] vars) {
			
			if (coeffs.length != vars.length)
				throw new IllegalArgumentException("Unexpected number of variables");
			
			double sum = 0;
			for (int i = 0; i < coeffs.length; ++i)
				sum += coeffs[i] * vars[i];
			return sum;
		}
		
	}


	private Map<String, VariableSet> varSets_ = new HashMap<String, VariableSet>();
	private int nextVarIdx_ = 0;
	
	private Collection<Constraint> constraints_ = new LinkedList<Constraint>();
	protected Collection<Constraint> getConstraints() { return constraints_; }
	
	private Expression objective_;
	protected Expression getObjective() { return objective_; }

	public VariableSet addVarSet(String name, int n1) {
		return addVarSet(name, n1, 1);
	}

	public VariableSet addVarSet(String name, int n1, int n2) {
		if (varSets_.containsKey(name))
			throw new IllegalArgumentException("Duplicate variable set name");
		
		VariableSet result = new VariableSet(name, nextVarIdx_, n1, n2);
		nextVarIdx_ += result.getVarCount();
		
		varSets_.put(name, result);
		return result;
	}
	
	public VariableSet getVarSet(String name) {
		return varSets_.get(name);
	}
	
	public int getVarCount() {
		return nextVarIdx_;
	}
	
	public void addConstraint(Constraint c) {
		constraints_.add(c);
	}
	
	public void setObjective(Expression value) {
		objective_ = value;
	}

	public double[] getConstVector(double value) {
		double[] result = new double[getVarCount()];
		for (int i = 0; i < result.length; ++i)
			result[i] = value;
		return result;
	}

}
