/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.hacks;

import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.peers.PeerManager;

public class DiscoveredPeersHack
{
	private static final Set<AnnouncePeer> EMPTY_ANNOUNCE_PEERS = Collections.unmodifiableSet(new HashSet<AnnouncePeer>());
	
	public static final byte DEFAULT_HANDSHAKE_TYPE = 0;
	public static final String DEFAULT_PEER_SOURCE = "<unknown>";
	
	public static final class AnnouncePeer {
		private String ip_;
		public String getIp() { return ip_; }
		
		private int tcpPort_;
		public int getTCPPort() { return tcpPort_; }
		
		private int udpPort_;
		public int getUDPPort() { return udpPort_; }

		public AnnouncePeer(PeerDescriptor peer) {
			ip_ = peer.getIP();
			tcpPort_ = peer.getTCPPort();
			udpPort_ = peer.getUDPPort();
		}

		public AnnouncePeer(String ip, int tcpPort, int udpPort) {
			ip_ = ip;
			tcpPort_ = tcpPort;
			udpPort_ = udpPort;
		}
		
		public boolean equals(Object o) {
			if (!(o instanceof AnnouncePeer))
				return false;
			
			AnnouncePeer oPeer = (AnnouncePeer)o;
			return tcpPort_ == oPeer.tcpPort_
				&& udpPort_ == oPeer.udpPort_
				&& ip_.equals(oPeer.ip_);
		}

		public int hashCode() {
			return ip_.hashCode();
		}
	}


	/* Class: PeerItem
	 * - Parameter used for addDiscoveredPeer
	 * - Result used in getDiscoveredPeer
	 */
	private static Class CLASS_PeerItem;
	private static Method METHD_PeerItem_getHandshakeType;
	private static Method METHD_PeerItem_getSource;
	private static Method METHD_PeerItem_convertSourceString;
	
	/* Class: PeerItem[]
	 * - Type for cached_peer_popularities (see below)
	 */
	private static Class CLASS_PeerItemArray;

	/* Class: PeerManagerImpl
	 * - Implements PeerManager
	 * - Provides access to PEPeerManager interface (which
	 *   is extended by the PEPeerControl interface) */
	private static Class CLASS_PeerManagerImpl;
	private static Field FIELD_PeerManagerImpl_manager;

	/* Class: PEPeerControlImpl
	 * - Implements PEPeerControl interface
	 * - Provides access to PeerDatabase */
	private static Class CLASS_PEPeerControlImpl;
	private static Field FIELD_PEPeerControlImpl_peer_database;

	/* Class: PeerDatabase
	 * - Provides access to addDiscoveredPeer
	 * - Provides access to discovered_peers (peers discovered from Tracker or DHT)
	 * - Provides access to cached_peer_popularities (peers discovered through PEX)
	 * - Provides access to popularity_pos (current index for peers discovered through PEX)
	 * - Provides access to map_mon for synchronization */
	private static Class CLASS_PeerDatabase;
	private static Method METHD_PeerDatabase_addDiscoveredPeer;
	private static Field FIELD_PeerDatabase_discovered_peers;
	private static Field FIELD_PeerDatabase_cached_peer_popularities;
	private static Field FIELD_PeerDatabase_popularity_pos;
	private static Field FIELD_PeerDatabase_map_mon;

	/* Class: AEMonitor
	 * - Provides sychronization for discovered_peers
	 */
	private static Class CLASS_AEMonitor;
	private static Method METHD_AEMonitor_enter;
	private static Method METHD_AEMonitor_exit;

	static {
		/*
		 * Initialize some items we need in order to get access
		 * to the discovered peer database.  We'll use some of Java's
		 * reflection and change some private/protected items
		 * of classes already loaded in the JVM to make them
		 * accessible by us.
		 */

		/*
		 * Locate the classes we need
		 */
		try {
			CLASS_PeerItem = Class.forName("com.aelitis.azureus.core.peermanager.peerdb.PeerItem");
			CLASS_PeerItemArray = Array.newInstance(CLASS_PeerItem, 0).getClass();
			METHD_PeerItem_getHandshakeType = CLASS_PeerItem.getDeclaredMethod("getHandshakeType", HackCommon.EMPTY_PARAMS);
			METHD_PeerItem_getSource = CLASS_PeerItem.getDeclaredMethod("getSource", HackCommon.EMPTY_PARAMS);
			METHD_PeerItem_convertSourceString = CLASS_PeerItem.getDeclaredMethod("convertSourceString", new Class[] { byte.class });
			
			CLASS_PeerManagerImpl = Class.forName("org.gudy.azureus2.pluginsimpl.local.peers.PeerManagerImpl");
			FIELD_PeerManagerImpl_manager = CLASS_PeerManagerImpl.getDeclaredField("manager");

			CLASS_PEPeerControlImpl = Class.forName("org.gudy.azureus2.core3.peer.impl.control.PEPeerControlImpl");
			FIELD_PEPeerControlImpl_peer_database = CLASS_PEPeerControlImpl.getDeclaredField("peer_database");

			CLASS_PeerDatabase = Class.forName("com.aelitis.azureus.core.peermanager.peerdb.PeerDatabase");
			METHD_PeerDatabase_addDiscoveredPeer = CLASS_PeerDatabase.getDeclaredMethod("addDiscoveredPeer",
				new Class[] {
					CLASS_PeerItem,
				});
			FIELD_PeerDatabase_discovered_peers = CLASS_PeerDatabase.getDeclaredField("discovered_peers");
			FIELD_PeerDatabase_cached_peer_popularities = CLASS_PeerDatabase.getDeclaredField("cached_peer_popularities");
			FIELD_PeerDatabase_popularity_pos = CLASS_PeerDatabase.getDeclaredField("popularity_pos");
			FIELD_PeerDatabase_map_mon = CLASS_PeerDatabase.getDeclaredField("map_mon");

			CLASS_AEMonitor = Class.forName("org.gudy.azureus2.core3.util.AEMonitor");
			METHD_AEMonitor_enter = CLASS_AEMonitor.getDeclaredMethod("enter", HackCommon.EMPTY_PARAMS);
			METHD_AEMonitor_exit = CLASS_AEMonitor.getDeclaredMethod("exit", HackCommon.EMPTY_PARAMS);
		} catch (Exception e) {
			throw new RuntimeException("Failed to locate required data through reflection", e);
		}

		/*
		 * Make the fields accessible
		 */
		FIELD_PeerManagerImpl_manager.setAccessible(true);
		FIELD_PEPeerControlImpl_peer_database.setAccessible(true);
		FIELD_PeerDatabase_discovered_peers.setAccessible(true);
		FIELD_PeerDatabase_cached_peer_popularities.setAccessible(true);
		FIELD_PeerDatabase_popularity_pos.setAccessible(true);
		FIELD_PeerDatabase_map_mon.setAccessible(true);
	}

	private static Object getPeerDatabase(PeerManager manager) throws Exception {

		/* Ensure manager is of correct type */
		if (!manager.getClass().isAssignableFrom(CLASS_PeerManagerImpl))
			throw new IllegalArgumentException();

		/* Get the PEPeerControl member */
		Object peerControl = FIELD_PeerManagerImpl_manager.get(manager);
		if (peerControl == null)
			throw new IllegalArgumentException();

		/* Ensure peerControl is of correct type */
		if (!peerControl.getClass().isAssignableFrom(CLASS_PEPeerControlImpl))
			throw new IllegalArgumentException();

		/* Get the PeerDatabase member */
		Object peerDatabase = FIELD_PEPeerControlImpl_peer_database.get(peerControl);
		if (peerDatabase == null)
			throw new IllegalArgumentException();

		return peerDatabase;
	}

	private static Object getPeerDatabaseMonitor(Object peerDatabase) throws Exception {
		Object peerDatabaseMonitor = FIELD_PeerDatabase_map_mon.get(peerDatabase);
		if (peerDatabaseMonitor == null)
			throw new IllegalArgumentException();
		return peerDatabaseMonitor;
	}

	private static void peerDatabaseMonitorEnter(Object peerDatabase) throws Exception {
		METHD_AEMonitor_enter.invoke(getPeerDatabaseMonitor(peerDatabase), HackCommon.EMPTY_ARGS);
	}

	private static void peerDatabaseMonitorExit(Object peerDatabase) throws Exception {
		METHD_AEMonitor_exit.invoke(getPeerDatabaseMonitor(peerDatabase), HackCommon.EMPTY_ARGS);
	}

	private static boolean checkPeerItem(Object peerItem) {
		return peerItem.getClass().isAssignableFrom(CLASS_PeerItem);
	}

	private static void checkPeerItemThrow(Object peerItem) throws Exception {
		if (!peerItem.getClass().isAssignableFrom(CLASS_PeerItem))
			throw new IllegalArgumentException();
	}

	private static void checkPeerItemThrow(Collection<PeerDescriptor> peerItems) throws Exception {
		for (PeerDescriptor peerItem : peerItems)
			checkPeerItemThrow(peerItem);
	}

	private static void addDiscoveredPeer(PeerManager manager, PeerDescriptor peerItem) throws Exception {
		checkPeerItemThrow(peerItem);

		/* Call the addDiscoveredPeer method; it performs synchronization itself */
		METHD_PeerDatabase_addDiscoveredPeer.invoke(getPeerDatabase(manager), new Object[] { peerItem });
	}

	private static List getDiscoveredPeersMember(Object peerDatabase) throws Exception {
		Object discoveredPeers = FIELD_PeerDatabase_discovered_peers.get(peerDatabase);
		if (discoveredPeers == null)
			throw new IllegalArgumentException();

		if (!(discoveredPeers instanceof List))
			throw new IllegalArgumentException();

		return (List)discoveredPeers;
	}

	/**
	 * Return and clear the current set of peers discovered via Peer Exchange (PEX).
	 * 
	 * @param manager PeerManager instance
	 * @return List of peers discovered
	 */
	public static List<PeerDescriptor> extractPEXPeers(PeerManager manager) {
		/* Retrieve and clear existing list of PEX peers.
		 * NOTE: There IS a race condition here since no monitor is used
		 *       to serialize access to this data structure.  However, the
		 *       worst thing that can happen is for us to retry a few PEX peers.
		 *       This happens in the case where we reset the cached_peer_popularities
		 *       field to null, the database rebuilds and 1 or more peers are selected,
		 *       and then the counter is reset to 0 (thus causing the next iteration
		 *       to retry the first couple of peers). This should not really matter,
		 *       however, since a Bloom filter is used to avoid connecting to the
		 *       same peer multiple times.
		 */
		
		try {
			Object peerDatabase = getPeerDatabase(manager);
			Object pexPeers = FIELD_PeerDatabase_cached_peer_popularities.get(peerDatabase);
			if (pexPeers == null)
				return Collections.emptyList();
	
			if (!(pexPeers.getClass().isAssignableFrom(CLASS_PeerItemArray)))
				throw new IllegalArgumentException();
	
			FIELD_PeerDatabase_cached_peer_popularities.set(peerDatabase, null);
			FIELD_PeerDatabase_popularity_pos.set(peerDatabase, 0);
	
			return Arrays.asList((PeerDescriptor[])pexPeers);
		} catch (Exception e) {
			throw new RuntimeException("Failed to extract PEX peers", e);
		}
	}

	/**
	 * Return the existing list of discovered peers
	 * 
	 * @param manager PeerManager instance
	 * @return List of discovered peers
	 */
	public static List<PeerDescriptor> getDiscoveredPeers(PeerManager manager) {
		try {
			Object peerDatabase = getPeerDatabase(manager);
			peerDatabaseMonitorEnter(peerDatabase);
			try {

				/* Get the current list */
				List discoveredPeers = getDiscoveredPeersMember(peerDatabase);

				/* Copy items to our own result list */
				List<PeerDescriptor> result = new ArrayList<PeerDescriptor>(discoveredPeers.size());
				for (Object peerItem : discoveredPeers) {
					checkPeerItemThrow(peerItem);
					result.add((PeerDescriptor)peerItem);
				}
				return result;
				
			} finally {
				peerDatabaseMonitorExit(peerDatabase);
			}
		} catch (Exception e) {
			throw new RuntimeException("Failed to get discovered peers", e);
		}
	}

	/**
	 * Return the size of the existing list of discovered peers
	 * 
	 * @param manager PeerManager instance
	 * @return Size of the list of discovered peers
	 */
	public static int getDiscoveredPeersCount(PeerManager manager) {
		try {
			Object peerDatabase = getPeerDatabase(manager);
			peerDatabaseMonitorEnter(peerDatabase);
			try {
				return getDiscoveredPeersMember(peerDatabase).size();
			} finally {
				peerDatabaseMonitorExit(peerDatabase);
			}
		} catch (Exception e) {
			throw new RuntimeException("Failed to get discovered peers", e);
		}
	}

	/**
	 * Replace the existing list of discovered peers.  Returns the set of discovered peers
	 * that was in the replaced list.
	 * 
	 * @param manager PeerManager instance
	 * @param peerItems New set of discovered peers
	 * @param targetPeers Filter the set of peers to be removed (if empty, all peers are removed)
	 * @return Previous set of discovered peers that were replaced
	 */
	public static List<PeerDescriptor> setDiscoveredPeers(PeerManager manager, List<PeerDescriptor> peerItems, Set<AnnouncePeer> targetPeers) {
		
		try {
			checkPeerItemThrow(peerItems);

			LinkedHashSet<PeerDescriptor> removedPeers = new LinkedHashSet<PeerDescriptor>();

			Object peerDatabase = getPeerDatabase(manager);
			peerDatabaseMonitorEnter(peerDatabase);
			try {

				/* Get the current list */
				List discoveredPeers = getDiscoveredPeersMember(peerDatabase);
				
				/* Remove all current items and put them in a hash set (PeerItem class
				 * implements proper hashCode method). Using a LinkedHashSet allows us
				 * to preserve the ordering. */
				for (Iterator peerItr = discoveredPeers.iterator(); peerItr.hasNext(); ) {
					Object peerItem = peerItr.next();
					checkPeerItemThrow(peerItem);
					PeerDescriptor peer = (PeerDescriptor)peerItem;

					/* Remove if targets is empty (i.e., we remove everything) or
					 * targets list contains the peer. */
					if (targetPeers.isEmpty() || targetPeers.contains(new AnnouncePeer(peer))) {
						removedPeers.add(peer);
						peerItr.remove();
					}
				}

				/* Add the new peers to the list, and remove duplicates from the set of removed peers */
				for (PeerDescriptor peerItem : peerItems) {
					removedPeers.remove(peerItem);
					addDiscoveredPeer(manager, peerItem);
				}

			} finally {
				peerDatabaseMonitorExit(peerDatabase);
			}
			
			/* Return an ordered list of removed peers */
			return new ArrayList<PeerDescriptor>(removedPeers);
		} catch (Exception e) {
			throw new RuntimeException("Failed to set discovered peers", e);
		}
	}
	
	public static List<PeerDescriptor> setDiscoveredPeers(PeerManager manager, List<PeerDescriptor> peerItems) {
		return setDiscoveredPeers(manager, peerItems, EMPTY_ANNOUNCE_PEERS);
	}

	public static byte getPeerHandshakeType(PeerDescriptor peer) {
		if (!checkPeerItem(peer))
			return DEFAULT_HANDSHAKE_TYPE;

		try {
			return (Byte)METHD_PeerItem_getHandshakeType.invoke(peer, HackCommon.EMPTY_ARGS);
		} catch (Exception e) {
			throw new RuntimeException("Failed to get peer's handshake type", e);
		}
	}

	public static String getPeerSource(PeerDescriptor peer) {
		if (!checkPeerItem(peer))
			return DEFAULT_PEER_SOURCE;
		
		try {
			byte sourceType = (Byte)METHD_PeerItem_getSource.invoke(peer, HackCommon.EMPTY_ARGS);
			return (String)METHD_PeerItem_convertSourceString.invoke(null, new Object[] { sourceType });
		} catch (Exception e) {
			throw new RuntimeException("Failed to get peer source type", e);
		}
	}

}
