/*
 * RoutingTable.java    1.0 2000/01/01
 *
 * Copyright (c) 2000 Stefan Nilsson
 * KTH, Nada
 * SE-100 44 Stockholm, Sweden
 * http://www.nada.kth.se/~snilsson
 *
 * The code presented in this file has been tested with care but
 * is not guaranteed for any purpose. The writer does not offer
 * any warranties nor does he accept any liabilities with respect
 * to the code.
 */

/*
 * Modifications made by Richard Alimi (richard.alimi@yale.edu)
 * for Vuze P4P Plugin developed by the Yale Laboratory of
 * Networked Systems.
 */

package se.kth.nada.router;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;

import net.yale.cs.p4p.api.data.PID;
import net.yale.cs.p4p.p2p.azureus.client.util.AddrUtils;

/**
 * A routing table for word sized (32 bits) bitstrings implemented as
 * a static level- and pathcompressed trie. For details please consult
 * <pre>
 *   Stefan Nilsson and Gunnar Karlsson. Fast Address Look-Up
 *   for Internet Routers. International Conference of Broadband
 *   Communications (BC'97).
 *
 *   http://www.nada.kth.se/~snilsson/papers/router/
 * </pre>
 *
 * @author  Stefan Nilsson
 * @version 1.0, 1 jan 2000
 */
public class RoutingTable {
    /**
     * The trie is represented by an array and each node consists of
     * an integer. The first 5 bits (31-27) indicate the logarithm
     * of the branching factor. The next 5 bits (26-22) indicate the
     * skip value. The final 22 (21-0) bits is an adress, either to
     * another internal node, or the base vector.
     * The maximum capacity is 2^21 strings (or a few more). The trie
     * is prefixfree. All strings that are prefixes of another string
     * are stored separately.
     */
    private int[] trie; // the trie in compact format
    private int nextfree; // global var used for mem alloc in trie array

    private ArrayList<Entry> entries;
    private Base[] base; // the base vector
    private Prefix[] pre; // the prefix vector
    private ArrayList<PID> nexthop; // the next-hop table

    private int rootBranch;
    private double fillFactor;

    private final static int ADR_SIZE = 32; // bits in an address
    private final static int FIRST_SKIP_BIT = 27;
    private final static int FIRST_ADR_BIT = 22;

    private final int getBranch(int w) {
        return w >>> FIRST_SKIP_BIT;
    }

    private final int getSkip(int w) {
        return w >>> FIRST_ADR_BIT & 0x1F;
    }

    private final int getAdr(int w) {
        return w & 0x3FFFFF;
    }

    private final static int NO_PREFIX = -1; // empty prefix pointer

    /**
     * Entries in the base vector are stored as a Base object
     */
    private static class Base {
        int str; // the routing entry
        int len; // and its length
        int pre; // pointer to prefix table, -1 if no prefix
        int nexthop; // pointer to next-hop table

        Base(int str, int len, int pre) {
            this.str = str;
            this.len = len;
            this.pre = pre;
        }
    }

    /**
     * Entries in the prefix vector are stored as a Prefix object
     */
    private static class Prefix {
        int len; // the length of the prefix
        int pre; // pointer to prefix, -1 if no prefix
        int nexthop; // pointer to nexthop table

        Prefix(int len, int pre) {
            this.len = len;
            this.pre = pre;
        }
    }

    /**
     * Extract n bits from str starting at position p.
     */
    private final int extract(int p, int n, int str) {
        return str << p >>> (ADR_SIZE - n);
    }

    /**
     * Remove the first p bits from str.
     */
    private final int remove(int p, int str) {
        return str << p >>> p;
    }

    /**
     * Constructs a routing table for the entries of the specified
     * array. The structure is packed as tightly as possible
     * (the fill factor is 1.0).
     *
     * @param entry   an array of entries to be stored in the table.
     * @param verbose   controls whether diagnostic messages are written
     *     to <tt>System.err</tt>.
     */
    public RoutingTable(ArrayList<Entry> entry) {
        this(entry, 1.0, 0);
    }

    private static ArrayList<Entry> readEntries(InputStream stream) throws IOException {
	    BufferedReader reader = new BufferedReader(new InputStreamReader(stream));

	    try {
	 	    ArrayList<Entry> entries = new ArrayList<Entry>();
		    while (true) {
			    String line = reader.readLine();
			    if (line == null)
				    break;

			    int sep1 = line.indexOf('\t');
			    if (sep1 < 0)
				    throw new IOException("Corrupt file");
		
			    int sep2 = line.indexOf('\t', sep1 + 1);
			    if (sep2 < 0)
				    throw new IOException("Corrupt file");
			    
			    entries.add(new Entry(
					    Integer.valueOf(line.substring(0, sep1)),
					    Integer.valueOf(line.substring(sep1 + 1, sep2)),
					    PID.parse(line.substring(sep2 + 1))));
		    }
		    
		    return entries;
	    } catch (NumberFormatException e) {
		    throw new IOException("Corrupt file");
	    }
    }

    public RoutingTable(InputStream stream) throws IOException {
	    this(readEntries(stream));
    }
    
    public void save(OutputStream stream) throws IOException {
	if (entries == null)
		throw new IOException("Attempted to save map twice");
	    
	PrintWriter writer = new PrintWriter(stream);
	try {
		for (Entry e : entries) {
	        	writer.append(Integer.toString(e.data));
	        	writer.append('\t');
	        	writer.append(Integer.toString(e.length));
	        	writer.append('\t');
	        	writer.append(e.nexthop.toString());
	        	writer.append('\n');
		}
	} finally {
		writer.flush();
	}
	
	// Only allow saving it once so we cut down on memory
	// during normal runtime
	entries = null;
    }

    /**
     * Constructs a routing table for the entries of the specified
     * array. By choosing a fill factor smaller than 1.0 it's possible
     * to gain speed at the expense of using a larger data structure.
     * Fixing the number of bits used for branching at the root of the
     * data structrure may also improve speed. 
     *
     * @param entry   an array of entries to be stored in the table.
     * @param fillFactor   an number indicating how tightly the data
     *     structure is packed (<tt>0.1 <= fillFactor <= 1.0</tt>).
     * @param rootBranch   a number indicating how many bits that are
     *     used for branching at the root of the data structure
     *     (<tt>0 <= rootBranch <= 20</tt>). The value 0 indicates that
     *     the branching is chosen dynamically.
     * @throws IllegalArgumentException if the fillFactor or rootBranch
     *         is out of range.
     */
    public RoutingTable(ArrayList<Entry> entry, double fillFactor, int rootBranch) {
        if (fillFactor < 0.1 || fillFactor > 1.0
                || rootBranch < 0 || rootBranch > 20)
            throw new IllegalArgumentException();

        int size; // Size of entry array after duplicate removal
        
        // handle 0-length prefix
        for (int i = 0; i < entry.size(); ++i) {
        	if (entry.get(i).length != 0)
        		continue;
        	
        	Entry e = entry.remove(i);
        	
        	// Replace with two 1-length prefixes
        	// FIXME: This hack depends on no existing 1-length prefixes
        	entry.add(new Entry(0x80000000, 1, e.nexthop));
        	entry.add(new Entry(0x00000000, 1, e.nexthop));
        }

        // save copy so we can save it later
        this.entries = new ArrayList<Entry>(entry);
        
        this.fillFactor = fillFactor;
        this.rootBranch = rootBranch;

        nexthop = buildNexthopTable(entry);

        Collections.sort(entry);
        // Remove duplicates
        size = entry.size() > 0 ? 1 : 0;
        for (int i = 1; i < entry.size(); i++)
            if (!entry.get(i-1).equals(entry.get(i)))
                entry.set(size++, entry.get(i));

        // We first build a big data structure and afterwards,
        // when we know the size, we store it in a more compact format.
        // The number of internal nodes in the tree can't be larger
        // than the number of entries and the number of leaves can
        // be at most size/fillFactor. If the branching is fixed at
        // the root there might be more leaves.
        int[] trie = new int[size + (int) (size/fillFactor) + (1<<rootBranch)];
        Base[] base = new Base[size];
        Prefix[] pre = new Prefix[size];

        // Initialize a temporary array of prefix pointers
        int[] preTemp = new int[size];
        for (int i = 0; i < size; i++)
            preTemp[i] = NO_PREFIX;

        // Go through the entries and put the prefixes in pre
        // and the rest of the strings in base
        int nbases = 0; // This many entries ended up in the base vector
        int nprefs = 0; // and this many in the prefix vector.
        for (int i = 0; i < size; i++)
            if (i < size-1 && entry.get(i).isPrefixOf(entry.get(i+1))) {
                Prefix ptemp = new Prefix(entry.get(i).length, preTemp[i]);
                // Update preTemp for all entries that have this prefix
                for (int j = i + 1;
                        j < size && entry.get(i).isPrefixOf(entry.get(j)); j++)
                    preTemp[j] = nprefs;
                ptemp.nexthop = binarySearch(entry.get(i).nexthop, nexthop);
                pre[nprefs++] = ptemp;
            } else {
                Base btemp = new Base(entry.get(i).data, entry.get(i).length, preTemp[i]);
                btemp.nexthop = binarySearch(entry.get(i).nexthop, nexthop);
                base[nbases++] = btemp;
            }

        // At this point we now how much memory to allocate
        // for the base vector and the prefix vector.
        this.base = new Base[nbases];
        this.pre = new Prefix[nprefs];
        for (int i = 0; i < nbases; i++)
            this.base[i] = base[i];
        for (int i = 0; i < nprefs; i++)
            this.pre[i] = pre[i];

        // Build the trie structure
        nextfree = 1;
        build(0, 0, nbases, 0, trie);

        // Allocate memory and move the tree into its final place.
        this.trie = new int[nextfree];
        for (int i = 0; i < nextfree; i++) {
            this.trie[i] = trie[i];
        }
    }

    /**
     * Compute the longest common prefix of the elements in the
     * sorted base array from position 'first' to 'first + n - 1'.
     * Skip the first 'prefix' characters.
     * We assume that n >= 2 and base[first] != base[first+n-1].
     */
    private int computeNewPrefix(int prefix, int first, int n) {
        int low = base[first].str;
        int high = base[first + n - 1].str;
        while (extract(prefix, 1, low) == extract(prefix, 1, high))
            prefix++;
        return prefix;
    }

    /**
     * Compute the branch value for the root of the tree that covers
     * the base array from position 'first' to 'first + n - 1'.
     * Disregard the first 'prefix' characters.
     * We assume that n >= 2 and base[first] != base[first+n-1].
     * Furthermore we assume that no bits should be skipped.
     */
    private int computeBranch(int prefix, int first, int n)
    {
        // Always use branching factor 2 for two elements
        if (n == 2)
            return 1;

        // Compute the number of bits b that can be used for branching.
        // We have at least two branches. Therefore we start the search
        // at b = 2, i.e. 2^b = 4 branches.
        int b = 1;
        int count;
        do {
             b++;
             if (n < fillFactor*(1<<b) ||
                 prefix + b > ADR_SIZE)
                 break;
             int i = first;
             count = 0;
             for (int pat = 0; pat < 1<<b; pat++) {
                 boolean patFound = false;
                 while (i < first + n &&   
                        pat == extract(prefix, b, base[i].str)) {
                     i++;
                     patFound = true;
                 }
                 if (patFound)
                     count++;
             }
        } while (count >= fillFactor*(1<<b));
        return b - 1;
    }

    /**
     * Compute the longest prefix match between bitpat and position p
     * in the base vector disregarding the first 'prefix' characters.
     */
    private int longestPrefixMatch(int prefix, int p,
                                   int bitpat, int branch) {
        int match = 0;
        int prep =  base[p].pre;
        while (prep != NO_PREFIX && match == 0) {
            int len = pre[prep].len;
            if (len > prefix &&
                  extract(prefix, len - prefix, base[p].str) ==
                  extract(ADR_SIZE - branch, len - prefix, bitpat))
                match = len;
            else
                prep = pre[prep].pre;
        }
        return match;
    }

    /**
     * Build a trie that covers the base array from position
     * 'first' to 'first + n - 1'. Disregard the first 'prefix'
     * characters. 'pos' is the position for the root of this
     * tree and 'nextfree' is the first position in the array
     * that hasn't yet been reserved.
     */
    private void build(int prefix, int first, int n,
                       int pos, int[] trie) {
        int branch, newprefix, adr;

        // only one element, this is a leaf
        if (n == 1) {
            trie[pos] = first; // branch and skip are 0
            return;
        }

        // compute skip value
        newprefix = computeNewPrefix(prefix, first, n);

        // compute branching
        if (rootBranch > 0 && prefix == 0 && first == 0)
            branch = rootBranch; // fixed branching at root
        else
            branch = computeBranch(newprefix, first, n);

        // adress of leftmost child
        adr = nextfree;
        trie[pos] = (branch << FIRST_SKIP_BIT) |
                    ((newprefix - prefix) << FIRST_ADR_BIT) |
                    adr;
        nextfree += 1<<branch;

        // Build the subtrees
        int p = first; // position in base array
        for (int bitpat = 0; bitpat < 1<<branch; bitpat++) {
            int k = 0; // count #entries matching this bit pattern    
            while (p + k < first + n &&
                   extract(newprefix, branch, base[p + k].str) == bitpat)
                k++;

            if (k == 0) {
                // The leaf should have a pointer either to p-1 or p,
                // whichever has the longest matching prefix
                int match1 = 0, match2 = 0;
                if (p > first)
                    match1 = longestPrefixMatch(newprefix, p - 1,
                                                bitpat, branch);
                if (p < first + n)
                    match2 = longestPrefixMatch(newprefix, p,
                                                bitpat, branch);

                if ((match1 > match2 && p > first) || p == first + n)
                    build(newprefix + branch, p - 1, 1,
                          adr + bitpat, trie);
                else
                    build(newprefix + branch, p, 1,
                          adr + bitpat, trie);

            } else if (k == 1 && base[p].len - newprefix < branch) {
                int bits = branch - base[p].len + newprefix;
                for (int i = bitpat; i < bitpat + (1<<bits); i++)
                    build(newprefix + branch, p, 1, adr + i, trie);
                bitpat += (1<<bits) - 1;

            } else
                build(newprefix + branch, p, k, adr + bitpat, trie);

            p += k;
        }
    }

    private ArrayList<PID> buildNexthopTable(ArrayList<Entry> entry) {
        // Extract the nexthop addresses from the entry array 
        ArrayList<PID> nextTemp = new ArrayList<PID>(entry.size());
        for (int i = 0; i < entry.size(); i++)
            nextTemp.add(entry.get(i).nexthop);
        
        // Sort it
        Collections.sort(nextTemp);

        // Remove duplicates
        int count = entry.size() > 0 ? 1 : 0;
        for (int i = 1; i < entry.size(); i++)
            if (!nextTemp.get(i-1).equals(nextTemp.get(i)))
                nextTemp.set(count++, nextTemp.get(i));

        // Move the elements to an array of proper size
        ArrayList<PID> nextHop = new ArrayList<PID>(count);
        for (int i = 0; i < count; i++) {
            nextHop.add(nextTemp.get(i));
        }
        return nextHop;
    }

    /**
     * Return the smallest i such that x == v[i].
     * If no such i exists, return -1.
     */
    private int binarySearch(PID x, ArrayList<PID> v) {
        int low, high, mid;

        low = 0;
        high = v.size() - 1;
        while (low <= high) {
            mid = (low + high) / 2;
            int compareRc = x.compareTo(v.get(mid));
            if (compareRc < 0)
                high = mid - 1;
            else if (compareRc > 0)
                low = mid + 1;
            else
                return mid;
        }
        return -1;
    }
    
    public PID find(String s) {
	    int addr = AddrUtils.addressToInt(s);
	    return (addr != AddrUtils.INVALID_ADDRESS) ? find(addr) : null;
    }

    /**
     * Performs a lookup.
     *
     * @param s   address to lookup.
     * @return nexthop  for address <tt>s</tt> or 0 if not found.
     */
    public PID find(int s) {
        // Traverse the trie
        int node = trie[0];
        int pos = getSkip(node);
        int branch = getBranch(node);
        int adr = getAdr(node);
        while (branch != 0) {
            node = trie[adr + extract(pos, branch, s)];
            pos += branch + getSkip(node);
            branch = getBranch(node);
            adr = getAdr(node);
        }

        // Was this a hit?
        int bitmask = base[adr].str ^ s;
        if (extract(0, base[adr].len, bitmask) == 0)
            return nexthop.get(base[adr].nexthop);

        // If not, look in the prefix tree
        int preadr = base[adr].pre;
        while (preadr != NO_PREFIX) {
            if (extract(0, pre[preadr].len, bitmask) == 0)
                return nexthop.get(pre[preadr].nexthop);
            preadr = pre[preadr].pre;
        }

        return null; // not found
    }

    /**
     * Prints the routing table to <tt>System.out</tt>.
     */
    public void print() {
        System.out.println();
        System.out.println("nexthops:");
        for (int i = 0;  i < nexthop.size(); i++)
            System.out.println("  " + i + " " + nexthop.get(i));

        System.out.println("prefix (len, pre, nexthop):");
        for (int i = 0;  i < pre.length; i++) {
            System.out.print("  " + i + " " + pre[i].len);
            if (pre[i].pre == NO_PREFIX)
                System.out.print(" E");
            else
                System.out.print(" " + pre[i].pre);
            System.out.println(" " + (pre[i].nexthop & 0xFFFFFFFFL));
        }

        System.out.println("base (pre, nexthop, len):");
        for (int i = 0;  i < base.length; i++) {
            System.out.print("  " + i);
            if (base[i].pre == NO_PREFIX)
                System.out.print(" E");
            else
                System.out.print(" " + base[i].pre);
            System.out.print(" " + (base[i].nexthop & 0xFFFFFFFFL));
            System.out.println(" " + base[i].len);
        }

        System.out.println("trie (branch, skip, adr):");
        for (int i = 0; i < trie.length; i++){
            int k = trie[i];
            System.out.print("  " + i + " " + getBranch(k));
            System.out.print(" " + getSkip(k));
            System.out.println(" " + getAdr(k));
        }
    }

    /**
     * Prints statistics about the data structure to <tt>System.out</tt>.
     *
     * @param verbose   controls if the output is verbose.
     */
    public void printStat(boolean verbose) {
        DecimalFormat df = new DecimalFormat("0.00");
        if (verbose) {
            System.err.println();
            System.err.println("Trie");
            System.err.print("  fill factor: ");
            System.err.println(df.format(fillFactor));
            System.err.print("  branch at root: ");
            if (rootBranch == 0)
                System.err.println(getBranch(trie[0]) + " bits");
            else
                System.err.println(rootBranch + " bits (fixed)");
            System.err.println("  size: " + 4L*trie.length + " bytes");
        }
        int intNodes = 0, leaves = 0;
        for (int i = 0; i < trie.length; i++)
            if (getBranch(trie[i]) == 0)
                leaves++;
            else
                intNodes++;
        if (verbose) {
            System.err.print("  " + leaves + " leaves, ");
            System.err.println(intNodes + " internal nodes");
        }

        int[] depth = new int[ADR_SIZE];
        traverse(trie[0], 0, depth);

        int totdepth = 0, max = 0;
        for (int i = 0; i < depth.length; i++) {
            totdepth += i*depth[i];
            if (depth[i] > 0)
                max = i;
        }
        if (verbose) {
            for (int i = 1; i <= max; i++)
                System.err.println("  " + i + " " + depth[i]);
            System.err.print("  aver path: ");
            System.err.println(df.format((double) totdepth / leaves));
            int maxB = 0;
            for (int i = 0; i < trie.length; i++) {
                int k  = trie[i];
                if (getBranch(k) > maxB)
                    maxB = getBranch(k);
            }
            System.err.println("  max branch: " + maxB + " bits");
        }

        if (!verbose) {
            System.err.print("fill: " + df.format(fillFactor));
            System.err.print(" root: ");
            if (rootBranch == 0)
                System.err.print(getBranch(trie[0]));
            else
                System.err.print(rootBranch);
            System.err.print(" aver: ");
            System.err.print(df.format((double) totdepth / leaves));
            System.err.print(" max: " + max);
            System.err.println(" size: " + 4L*trie.length);
        }

        if (verbose) {
            System.err.println("Base vector");
            System.err.println("  " + base.length + " entries");
            System.err.println("Prefix vector");
            System.err.println("  " + pre.length + " entries");

            int maxP = 0, aver = 0;
            for (int i = 0; i < pre.length; i++) {
                Prefix p = pre[i];
                int j;
                for (j = 0; p.pre != NO_PREFIX; j++)
                    p = pre[p.pre];
                if (j > maxP)
                    maxP = j;
                aver += j;
            }
            System.err.print("  aver path: ");
            System.err.println(df.format((double) aver / pre.length));
            System.err.println("  max path: " + maxP);

            System.err.println("Nexthop vector");
            System.err.println("  " + nexthop.size() + " entries");
            System.err.println();
        }
    }

    /**
     * Traverse the trie and count depths of all leaves.
     */
    private void traverse(int root, int depth, int depths[]) {
        if (getBranch(root) == 0) // leaf
            depths[depth]++;
        else
            for (int i = 0; i < 1<<getBranch(root); i++)
                traverse(trie[getAdr(root)+i], depth + 1, depths);
    }
}
